#include "SceneNode.h"

#include "Command.h"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <iostream>

using namespace DirectX;

SceneNode::SceneNode(Category::Type category)
	: mTrasforms()
	, mChildren()
	, mParent(nullptr)
	, mDefaultCategory(category)
{
	XMStoreFloat4x4(&mWorld, XMMatrixIdentity());
}

void SceneNode::attachChild(Ptr child)
{
	child->mParent = this;
	mChildren.push_back(std::move(child));
}

SceneNode::Ptr SceneNode::detachChild(const SceneNode& node)
{
	auto found = std::find_if(mChildren.begin(), mChildren.end(), [&] (Ptr& p) { return p.get() == &node; });
	assert(found != mChildren.end());

	Ptr result = std::move(*found);
	result->mParent = nullptr;
	mChildren.erase(found);
	return result;
}

void SceneNode::Update(float dt, CommandQueue& commands)
{
	updateCurrent(dt, commands);
	updateChildren(dt, commands);
}

void SceneNode::checkSceneCollision(SceneNode & sceneGraph, std::set<Pair>& collisionPairs)
{
	checkNodeCollision(sceneGraph, collisionPairs);

	for (Ptr& child : sceneGraph.mChildren)
		checkSceneCollision(*child, collisionPairs);
}

void SceneNode::checkNodeCollision(SceneNode & node, std::set<Pair>& collisionPairs)
{
	if (this != &node && collision(*this, node) && !isDestroyed() && !node.isDestroyed())
		collisionPairs.insert(std::minmax(this, &node));

	for (Ptr& child : mChildren)
		child->checkNodeCollision(node, collisionPairs);
}

void SceneNode::removeWrecks()
{
	// Remove all children which request so
	auto wreckfieldBegin = std::remove_if(mChildren.begin(), mChildren.end(), std::mem_fn(&SceneNode::isMarkedForRemoval));
	mChildren.erase(wreckfieldBegin, mChildren.end());

	// Call function recursively for all remaining children
	std::for_each(mChildren.begin(), mChildren.end(), std::mem_fn(&SceneNode::removeWrecks));
}

DirectX::BoundingOrientedBox SceneNode::getBoundingRect() const
{
	return DirectX::BoundingOrientedBox();
}

bool SceneNode::isMarkedForRemoval() const
{
	// By default, remove node if entity is destroyed
	return isDestroyed();
}

bool SceneNode::isDestroyed() const
{
	// By default, scene node needn't be removed
	return false;
}

void SceneNode::updateCurrent(float dt, CommandQueue& commands)
{
	// Do nothing by default
}

void SceneNode::updateChildren(float dt, CommandQueue& commands)
{
	for (Ptr& child : mChildren)
		child->Update(dt, commands);
}

void SceneNode::Draw(ID3D11DeviceContext* dc, const Camera& camera, const CommonStates& states) const
{
	// Apply transform of current node
	// TODO - poprawic transforamcje tak jak powinny byc // zrobic stutct 3x XMVector ???
	//states.transform *= getTransform();

	// Draw node and children with changed transform
	drawCurrent(dc, camera, states);
	drawChildren(dc, camera, states);

	// Draw bounding rectangle - disabled by default
	//drawBoundingRect(dc, camera, states);
}

void SceneNode::setPosition(DirectX::XMFLOAT3 position)
{
	mTrasforms.mTranslate = position;
}

void SceneNode::setPosition(float x, float y, float z)
{
	mTrasforms.mTranslate.x = x;
	mTrasforms.mTranslate.y = y;
	mTrasforms.mTranslate.z = z;
}

void SceneNode::drawCurrent(ID3D11DeviceContext* deviceContext, const Camera& camera, const DirectX::CommonStates& states) const
{
	// Do nothing by default
}

void SceneNode::drawChildren(ID3D11DeviceContext* deviceContext, const Camera& camera, const DirectX::CommonStates& states) const
{
	for (const Ptr& child : mChildren)
		child->Draw(deviceContext, camera, states);
}

void SceneNode::drawBoundingRect(ID3D11DeviceContext * deviceContext, const Camera & camera, const DirectX::CommonStates & states) const
{
	// TODO - dorobiz z DirectXTK przykadu collisions
}

Transforms & SceneNode::getTransform()
{
	return mTrasforms;
}

void SceneNode::setWorld(DirectX::FXMMATRIX& world)
{
	XMStoreFloat4x4(&mWorld, world);
}

void SceneNode::setWorld(DirectX::XMFLOAT4X4& world)
{
	mWorld = world;
}


const Transforms & SceneNode::getTransform() const
{
	return mTrasforms;
}

DirectX::FXMMATRIX SceneNode::getWorld() const
{
	return mTrasforms.World() * XMLoadFloat4x4(&this->mWorld);
}

DirectX::FXMVECTOR SceneNode::getWorldPosition() const
{
	// TODO - sprawdzic czy OK.
	return XMVector3TransformCoord(XMLoadFloat3(&XMFLOAT3(1.0f, 1.0f, 1.0f)), getWorldTransform());
	//return getWorldTransform() * XMLoadFloat3(&XMFLOAT3(1.0f, 1.0f, 1.0f));
}

DirectX::FXMMATRIX SceneNode::getWorldTransform() const
{
	XMMATRIX transform = XMMatrixIdentity();

	const SceneNode* node = this;
	return node->mParent->getWorld();

	/*for (const SceneNode* node = this; node != nullptr; node = node->mParent)
		transform = node->getTransform().World() * transform;

	return transform;*/
}

DirectX::FXMMATRIX SceneNode::getWorldWorlds() const
{
	XMMATRIX transform = XMMatrixIdentity();

	for (const SceneNode* node = this; node != nullptr; node = node->mParent)
		transform = XMLoadFloat4x4(&node->mWorld) *  transform;

	return transform;
}

void SceneNode::onCommand(const Command& command, float dt)
{
	// Command current node, if category matches
	if (command.category & getCategory())
		command.action(*this, dt);

	// Command children
	for (Ptr& child : mChildren)
		child->onCommand(command, dt);
}

unsigned int SceneNode::getCategory() const
{
	return mDefaultCategory;
}

bool collision(const SceneNode & lhs, const SceneNode & rhs)
{
	return lhs.getBoundingRect().Intersects(rhs.getBoundingRect());
}

float distance(const SceneNode& lhs, const SceneNode& rhs)
{
	// TODO = chyba OK
	return XMVectorGetX(XMVector3Length(lhs.getWorldPosition() - rhs.getWorldPosition()));
}