#ifndef BOOK_SCENENODE_HPP
#define BOOK_SCENENODE_HPP

#include "Category.h"
#include "CommandQueue.h"

#include <vector>
#include <memory>
#include <set>

#include <d3d11.h>
#include <DirectXMath.h>
#include <CommonStates.h>
#include <DirectXCollision.h>

#include <DebugDraw.h>
#include <Effects.h>
#include <PrimitiveBatch.h>
#include <SpriteBatch.h>

#include "Camera.h"
#include "Transforms.h"

extern std::unique_ptr<DirectX::BasicEffect> g_debugEffect;
extern std::unique_ptr<DirectX::PrimitiveBatch<DirectX::VertexPositionColor>> g_debugBatch;
extern std::unique_ptr<ID3D11InputLayout> g_debugLayout;
extern std::unique_ptr<DirectX::SpriteBatch> g_spriteBatch;

struct Command;

struct FloatRect
{
	float x;
	float y;
	float w;
	float h;

	FloatRect() = default;
	FloatRect(float _x, float _y, float _w, float _h) : x(_x), y(_y), w(_w), h(_h) {}
};

class SceneNode
{
public:
	typedef std::unique_ptr<SceneNode> Ptr;
	typedef std::pair<SceneNode*, SceneNode*> Pair;

public:
	SceneNode(Category::Type category = Category::None);
	virtual ~SceneNode() = default;

	void attachChild(Ptr child);
	Ptr detachChild(const SceneNode& node);
		
	virtual void Update(float dt, CommandQueue& commands);
	virtual void Draw(ID3D11DeviceContext* dc, const Camera& camera, const CommonStates& states) const;

	void setPosition(DirectX::XMFLOAT3 position);
	void setPosition(float x, float y, float z);

	Transforms& getTransform();
	const Transforms& getTransform() const;
	DirectX::FXMMATRIX getWorld() const;
	DirectX::FXMVECTOR getWorldPosition() const;
	DirectX::FXMMATRIX getWorldTransform() const;
	DirectX::FXMMATRIX getWorldWorlds() const;

	void setWorld(DirectX::FXMMATRIX& world);
	void setWorld(DirectX::XMFLOAT4X4& world);

	void onCommand(const Command& command, float dt);
	virtual unsigned int getCategory() const;

	void checkSceneCollision(SceneNode& sceneGraph, std::set<Pair>& collisionPairs);
	void checkNodeCollision(SceneNode& node, std::set<Pair>& collisionPairs);
	void removeWrecks();
	virtual DirectX::BoundingOrientedBox getBoundingRect() const;
	virtual bool isMarkedForRemoval() const;
	virtual bool isDestroyed() const;

	Transforms mTrasforms;
protected:
	virtual void updateCurrent(float dt, CommandQueue& commands);
	void updateChildren(float dt, CommandQueue& commands);

	virtual void drawCurrent(ID3D11DeviceContext* deviceContext, const Camera& camera, const DirectX::CommonStates& states) const;
	void drawChildren(ID3D11DeviceContext* deviceContext, const Camera& camera, const DirectX::CommonStates& states) const;
	virtual void drawBoundingRect(ID3D11DeviceContext* deviceContext, const Camera& camera, const DirectX::CommonStates& states) const;


protected:
	
	DirectX::XMFLOAT4X4 mWorld;
	DirectX::BoundingOrientedBox mBoundingOrientedBox;

private:
	std::vector<Ptr> mChildren;
	SceneNode* mParent;
	Category::Type mDefaultCategory;

	friend class World;
};

bool collision(const SceneNode& lhs, const SceneNode& rhs);
float distance(const SceneNode& lhs, const SceneNode& rhs);

#endif // BOOK_SCENENODE_HPP
