﻿#include "SnakeApp.h"

using namespace DirectX;

std::unique_ptr<DirectX::BasicEffect> g_debugEffect;
std::unique_ptr<DirectX::PrimitiveBatch<DirectX::VertexPositionColor>> g_debugBatch;
std::unique_ptr<ID3D11InputLayout> g_debugLayout;
std::unique_ptr<DirectX::SpriteBatch> g_spriteBatch;

SnakeApp::SnakeApp(HINSTANCE hInstance)
	: D3DApp(hInstance)
{
	mMainWndCaption = L"Blend Demo";
	mEnable4xMsaa = false;

	mLastMousePos.x = 0;
	mLastMousePos.y = 0;

	XMMATRIX I = XMMatrixIdentity();
}

SnakeApp::~SnakeApp()
{
	md3dImmediateContext->ClearState();
}

bool SnakeApp::Init()
{
	if(!D3DApp::Init())
		return false;

	states = std::make_unique<CommonStates>(md3dDevice);
	m_fxFactory = std::make_unique<EffectFactory>(md3dDevice);
	m_spriteBatch = std::make_unique<SpriteBatch>(md3dImmediateContext);
	mModels = std::make_unique<ModelHolder>(md3dDevice, *m_fxFactory);
	loadModels();

	// Init global debug objects
	{
		g_debugBatch = std::make_unique<PrimitiveBatch<VertexPositionColor>>(md3dImmediateContext);
		g_debugEffect = std::make_unique<BasicEffect>(md3dDevice);
		g_debugEffect->SetVertexColorEnabled(true);
		g_spriteBatch = std::make_unique<SpriteBatch>(md3dImmediateContext);

		void const* shaderByteCode;
		size_t byteCodeLength;

		g_debugEffect->GetVertexShaderBytecode(&shaderByteCode, &byteCodeLength);

		ID3D11InputLayout *pInputLayout{};
		md3dDevice->CreateInputLayout(
			VertexPositionColor::InputElements, VertexPositionColor::InputElementCount,
			shaderByteCode, byteCodeLength,
			&pInputLayout);
		g_debugLayout = std::unique_ptr<ID3D11InputLayout>(pInputLayout);
	}

	mContext = std::make_unique<Context>(*this, md3dDevice, md3dImmediateContext, *mModels, mPlayer, *m_spriteBatch, *states);

	mWorld = std::make_unique<World>(*mContext);	 

	return true;
}

void SnakeApp::OnResize()
{
	D3DApp::OnResize();

	if (mWorld.get() == nullptr)
		return;

	mWorld->getCamera().SetLens(0.25f*MathHelper::Pi, AspectRatio(), 1.0f, 1000.0f);
	//mCam.SetLens(0.25f*MathHelper::Pi, AspectRatio(), 1.0f, 1000.0f);
}

void SnakeApp::UpdateScene(float dt)
{
	mWorld->Update(dt);

	// Zamień współrzędne sferyczne na kartezjańskie.
	//float x = mRadius*sinf(mPhi)*cosf(mTheta);
	//float z = mRadius*sinf(mPhi)*sinf(mTheta);
	//float y = mRadius*cosf(mPhi);

	//
	// Control the camera.
	//
	if (GetAsyncKeyState('W') & 0x8000)
		mWorld->getCamera().Walk(10.0f*dt);

	if (GetAsyncKeyState('S') & 0x8000)
		mWorld->getCamera().Walk(-10.0f*dt);

	if (GetAsyncKeyState('A') & 0x8000)
		mWorld->getCamera().Strafe(-10.0f*dt);

	if (GetAsyncKeyState('D') & 0x8000)
		mWorld->getCamera().Strafe(10.0f*dt);
}

void SnakeApp::DrawScene()
{
	md3dImmediateContext->ClearRenderTargetView(mRenderTargetView, reinterpret_cast<const float*>(&Colors::Cyan));
	md3dImmediateContext->ClearDepthStencilView(mDepthStencilView, D3D11_CLEAR_DEPTH|D3D11_CLEAR_STENCIL, 1.0f, 0);

	mWorld->Draw();

	HR(mSwapChain->Present(0, 0));
}

LRESULT SnakeApp::MsgProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	CommandQueue& commands = mWorld->getCommandQueue();

	// update keyboards state
	mPlayer.WndProc(hwnd, msg, wParam, lParam);

	mPlayer.handleEvent(commands);
	mPlayer.handleRealtimeInput(commands);

	return D3DApp::MsgProc(hwnd, msg, wParam, lParam);
}

void SnakeApp::OnMouseDown(WPARAM btnState, int x, int y)
{
	mLastMousePos.x = x;
	mLastMousePos.y = y;

	SetCapture(mhMainWnd);
}

void SnakeApp::OnMouseUp(WPARAM btnState, int x, int y)
{
	ReleaseCapture();
}

void SnakeApp::OnMouseMove(WPARAM btnState, int x, int y)
{
	if ((btnState & MK_LBUTTON) != 0)
	{
		// Make each pixel correspond to a quarter of a degree.
		float dx = XMConvertToRadians(0.25f*static_cast<float>(x - mLastMousePos.x));
		float dy = XMConvertToRadians(0.25f*static_cast<float>(y - mLastMousePos.y));

		mWorld->getCamera().Pitch(dy);
		mWorld->getCamera().RotateY(dx);
	}

	mLastMousePos.x = x;
	mLastMousePos.y = y;
}

CommonStates & SnakeApp::getCommonStates() const
{
	return *states;
}

DirectX::IEffectFactory* SnakeApp::getEffectFactory() const
{
	return m_fxFactory.get();
}

void SnakeApp::loadModels()
{
	
	mModels->load(Models::GenericPalm, L"GenericPalm.cmo");
	mModels->load(Models::SnakeHead, L"SnakeBodyHead.cmo");
	mModels->load(Models::SnakeBody, L"SnakeBody.cmo");
	mModels->load(Models::Banana, L"appleWithSteam.cmo");
	mModels->load(Models::Apple, L"banana.cmo");
	mModels->load(Models::ExtraLive, L"extraLive.cmo");

}