#include "VideoSource.h"

#include <opencv2\core.hpp>
#include <opencv2\core\directx.hpp>
#include <opencv2\imgproc.hpp>

using namespace std;

VideoSource::VideoSource(ID3D11Device* device, ID3D11DeviceContext* immediateContext)
	: md3dDevice(device)
	, md3dImmediateContext(immediateContext)
	, mVideoPlayThread(nullptr)
	, mIsRunningThread(false)
	, mTryReopen(false)
	, mIsNotRetrivedFrame(false)
	, mIsPlaying(false)
	, m_pSurfaceRGBA(nullptr)
	, mIsActualSRV(false)
	, mFPS(0)
{
}

VideoSource::VideoSource(ID3D11Device * device, ID3D11DeviceContext* immediateContext, int camera_index)
	: VideoSource(device, immediateContext)
{
	mCamIndex = camera_index;
}

VideoSource::VideoSource(ID3D11Device * device, ID3D11DeviceContext* immediateContext, std::string source)
	: VideoSource(device, immediateContext)
{
	mCamIndex = -1;
	mSource = source;
}

VideoSource::~VideoSource()
{
	if (mIsRunningThread) {
		//this->mFrameMutex.lock();
		ReleasePlayVideoThread();
		//this->mFrameMutex.unlock();
	}
}

void VideoSource::Play()
{
	OpenVideo();

	mFPS = mVideoCap.get(CV_CAP_PROP_FPS);

	mVideoPlayThread = new std::thread(&VideoSource::PlayVideoThread, this);
}

void VideoSource::Stop()
{
	ReleasePlayVideoThread();
}

bool VideoSource::IsPlaying()
{
	return mIsPlaying;
}

bool VideoSource::IsNotRetrivedFrame()
{
	return mIsNotRetrivedFrame;
}

const cv::Mat VideoSource::GetAsMat() const
{
	measureTime("lock");
	mFrameMutex.lock();
	//logTime();
	measureTime("clone");
	cv::Mat frame = mCurrFrame.clone();
	//logTime();
	measureTime("unlock");
	mIsNotRetrivedFrame = false;
	mFrameMutex.unlock();
	//logTime();
	return frame;
}

ID3D11Texture2D* VideoSource::GetAsD3D11Texture2D()
{
	mFrameMutex.lock();
	cv::Mat frame = mCurrFrame.clone();
	mIsNotRetrivedFrame = false;
	mFrameMutex.unlock();

	vector<cv::Mat> channels;
	cv::Mat alphaImage;//(frame.rows, frame.cols, CV_8UC4);
	cv::split(frame, channels);
	channels.push_back(cv::Mat(frame.rows, frame.cols, CV_8UC1, cv::Scalar(255)));
	cv::merge(channels, alphaImage);

	//ID3D11Texture2D* d3d11Texture = this->CreateTexture(frame.cols, frame.rows);
	ID3D11Texture2D* surface = nullptr;
	get_surface(&surface);
	//cv::directx::convertToD3D11Texture2D(frame, d3d11Texture);

	return surface;
}

ID3D11ShaderResourceView * VideoSource::LastFrameSRV()
{
	std::lock_guard<std::mutex> guard(mLastSRVMutex);
	if (mIsActualSRV)
		return mLastSRV.get();

	D3D11_TEXTURE2D_DESC desc;
	ID3D11Texture2D *pTexture2D;
	get_surface(&pTexture2D);
	//pTexture2D = GetAsD3D11Texture2D();
	pTexture2D->GetDesc(&desc);

	D3D11_SHADER_RESOURCE_VIEW_DESC srv_dc;
	srv_dc.Format = desc.Format;
	srv_dc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
	srv_dc.Texture2D.MipLevels = desc.MipLevels;
	srv_dc.Texture2D.MostDetailedMip = desc.MipLevels - 1;
	// TO DO
	// byc moze trzeba tu dac jakis lock/mutex ma mLastSRV
	if (mLastSRV) {
		mLastSRV.release();
		//mLastSRV = nullptr;
	}
	ID3D11ShaderResourceView *ppSRView;
	md3dDevice->CreateShaderResourceView(pTexture2D, &srv_dc, &ppSRView);
	mLastSRV = std::move(std::unique_ptr<ID3D11ShaderResourceView>(ppSRView));

	mIsActualSRV = true;
	return mLastSRV.get();
}

void VideoSource::OpenVideo()
{
	std::wstringstream wstr;
	wstr << L"Cannot open a video camara: ";

	 if (mCamIndex >= 0) {
		 wstr << mCamIndex;
		 mVideoCap.open(mCamIndex);
	 }
	 else {
		 wstr << mSource.c_str();
		 mVideoCap.open(mSource);
	 }

	if (!mVideoCap.isOpened()) {
		MessageBox(nullptr, wstr.str().c_str(), L"info", 0);

		std::runtime_error("Cannot open a video camara");
	}
	//else
		//MessageBox(nullptr, L"otworto zrodlo wideo", L"info", 0);
}

void VideoSource::ReleasePlayVideoThread()
{
	if (mVideoPlayThread)
	{
		if (mIsRunningThread)
			mIsRunningThread = false;
		mVideoPlayThread->join();
		delete mVideoPlayThread;
		mVideoPlayThread = nullptr;
		mIsPlaying = false;
	}
}

void VideoSource::PlayVideoThread()
{
	assert(mVideoCap.isOpened());
	mIsRunningThread = true;

	while (mIsRunningThread)
	{
		std::this_thread::sleep_for(std::chrono::milliseconds(1000 / mFPS));
		if (mVideoCap.grab())
		{
			cv::Mat frame;
			if (mVideoCap.retrieve(frame))
			{
				mIsPlaying = true;
				mFrameMutex.lock();
				cv::flip(frame, mCurrFrame, 1);
				//mCurrFrame = frame;
				if (m_pSurfaceRGBA == nullptr)
					m_pSurfaceRGBA = this->CreateTexture(mCurrFrame.cols, mCurrFrame.rows);
				//cv::imshow("cam", mCurrFrame);
				//cv::waitKey(5);
				mIsNotRetrivedFrame = true;
				mIsActualSRV = false;
				mFrameMutex.unlock();
			}
			else
			{
				mIsPlaying = false;
				mTryReopen = true;
				//mIsRunningThread = false;
			}
		}
	}
}

std::unique_ptr<ID3D11Texture2D> VideoSource::CreateTexture(int width, int height)
{
	D3D11_TEXTURE2D_DESC desc;
	desc.Width = width;
	desc.Height = height;
	desc.MipLevels = desc.ArraySize = 1;
	desc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	desc.SampleDesc.Count = 1;
	desc.SampleDesc.Quality = 0;
	desc.Usage = D3D11_USAGE_DYNAMIC;
	desc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
	desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	desc.MiscFlags = 0;

	ID3D11Texture2D *pTexture = nullptr;
	HR(md3dDevice->CreateTexture2D(&desc, nullptr, &pTexture));

	return std::move(std::unique_ptr<ID3D11Texture2D>(pTexture));
}

void VideoSource::get_surface(ID3D11Texture2D ** ppSurface)
{
	HRESULT r;

	std::lock_guard<std::mutex> guard(mFrameMutex);

	if (mCurrFrame.empty())
		return;

	cv::Mat frame_rgba;
	cv::cvtColor(mCurrFrame, frame_rgba, CV_BGR2RGBA);
	cv::flip(frame_rgba, frame_rgba, 1);

	// process video frame on CPU
	UINT subResource = ::D3D11CalcSubresource(0, 0, 1);

	D3D11_MAPPED_SUBRESOURCE mappedTex;
	r = md3dImmediateContext->Map(m_pSurfaceRGBA.get(), subResource, D3D11_MAP_WRITE_DISCARD, 0, &mappedTex);
	if (FAILED(r))
	{
		throw std::runtime_error("surface mapping failed!");
	}

	cv::Mat m(mCurrFrame.rows, mCurrFrame.cols, CV_8UC4, mappedTex.pData, mappedTex.RowPitch);
	frame_rgba.copyTo(m);

	md3dImmediateContext->Unmap(m_pSurfaceRGBA.get(), subResource);


	*ppSurface = m_pSurfaceRGBA.get();
} // get_surface()